"""
# S2(x,y) correction codes
**Jianyu Long (jylong@uchicago.edu)**

```python
__version__='4.1.0'
```

Position dependent electron extraction efficiency, secondary scintillation gain and light detection efficiency, S2 signals in LXeTPC has a large position dependence. In order to make a precise energy reconstruction, we need a “correction” map to overcome its spatial dependence. This usually requires a selection of S2s originated from the same number of electrons, and then derive its average value.

See [description in the Team C overview page](https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:analysis:reconstruction_team#s2_x_y_correction)

## Brief info

- The map is created based on a crude Kr selection by default, if you want to investigate more monoenergetic sources, follow the instruction below
- The script contains a Cartesian map and a Polar map. They are independent from the other. You can inherite from either one for your purpose.
- To use, you must have an output from straxen/event_info about electron lifetime correction coefficient (it is used in .do_elifetime_correct)


"""

from ..base_references import BaseMap


class S2XYMap(BaseMap):
    _ALIAS = "s2_xy_maps"



"""
# Correction: Relative charge yield
# What is it correction: Corrected areas (cS2)

# Description: Correction done to the area of cS2s in our detector. The relative charge yield corresponds to the number of electrons released by an interaction per energy. This correction helps us convert from the number of electrons observed in an S2 event to an energy.

# Latest wiki reference: https://xe1t-wiki.lngs.infn.it/doku.php?id=caio:analysis:sr2:v18.5_relative_cy_map

"""


import rframe

from ..base_corrections import TimeSampledCorrection


class RelativeChargeYield(TimeSampledCorrection):

    _ALIAS = "relative_charge_yield"
    value: float



from typing import Literal
from .base_report import BaseOperationsReport

class DetectorOperations(BaseOperationsReport):
    """ Detector Operations Reports """

    _ALIAS="detector_operations"

    #system:Literal["CRY", "PUR", "LXePUR", "RSX", "RSX_2", "DST", "RAD", "DAQ"]
    system: str = ""
    label: str = ""
    details: str = ""
    link: str = ""




import os
import fsspec
from pydantic import BaseSettings, root_validator
import yaml

from ..utils import Database, LazyDataAccessor
from .data_folder import DataFolder



def git_config_source(cls, **kwargs):
    """
    Get github credentials from git config
    """
    try:
        from git import config

        cfg_path = config.get_config_path("global")
        cfg = config.GitConfigParser(cfg_path)
        return dict(username=cfg.get("github", "user"),
                    token=cfg.get("github", "token"))
    except:
        return {}


class GithubCredentials(BaseSettings):
    class Config:
        env_prefix = "github_"
        secrets_dir = '/run/secrets'

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):

            sources = (
                init_settings,
                env_settings,
                git_config_source,
            )
            if os.path.isdir("/run/secrets"):
                sources = sources + (file_secret_settings,)
            return sources


    username: str = "__token__"
    token: str = None

    @classmethod
    def find(cls, **kwargs):
        inst = cls(**kwargs)
        if inst.token is None:
            inst = cls.from_login()
        if inst.token is None:
            raise RuntimeError("Could not find github credentials.")
        return inst

    @classmethod
    def from_login(cls):
        from xeauth.github import GitHubAuth
        auth = GitHubAuth.device_login(scopes=["read:org", "read:user", "repo"])
        return cls(username=auth.api.username, token=auth.oauth_token)


class GithubRepo(DataFolder):
    class Config:
        env_prefix = "XEDOCS_GITHUB_REPO_"

    protocol: str = "github"
    org: str = "XENONnT"
    repo: str = "xedocs-data"
    username: str = "__token__"
    token: str = None
    branch: str = None
    
    def abs_path(self, path):
        if isinstance(path, list):
            return [self.abs_path(p) for p in path]
        return f"github://{path.lstrip('/')}"

    def storage_kwargs(self, path):
        return {
            "org": self.org,
            "repo": self.repo,
            "username": self.username,
            "token": self.token,
            "sha": self.branch,
        }

    @root_validator
    def get_token(cls, values):
        token = values.get('token', None)
        if token is None:
            cred = GithubCredentials.find()
            values['username'] = cred.username
            values['token'] = cred.token
        return values
    
    @property
    def datasets_config(self):
        return self.read_config()




import os
import yaml
import fsspec

from pydantic import BaseSettings, validator

from ..utils import Database, LazyFileAccessor


class DataFolder(BaseSettings):
    
    class Config:
        env_prefix = "XEDOCS_DATA_FOLDER_"

    root: str = None
    protocol: str = "file"
    config_path: str = "datasets.yml"

    @validator("root", pre=True)
    def expand_user(cls, value):
        if isinstance(value, str):
            value = os.path.expanduser(value)
        return value

    def storage_kwargs(self, path):
        return {}

    def abs_path(self, path):
        if isinstance(path, list):
            return [self.abs_path(p) for p in path]
        path = os.path.expanduser(path)
        if not os.path.isabs(path) and self.root is not None:
            path = os.path.join(self.root, path)
        return path

    @property
    def datasets_config(self):
        return self.read_config()

    def get_datasets(self):
        import xedocs

        dsets = {}
        for _, cfg in self.datasets_config.items():
            schema = cfg.get('schema', None)
            if schema is None:
                continue
            schema = xedocs.find_schema(schema)
            path = cfg.get('path', None)
            if path is None:
                continue
            path = self.abs_path(path)

            kwargs = self.storage_kwargs(path)
            dsets[schema._ALIAS] = LazyFileAccessor(schema,
                                                    path, 
                                                    **kwargs)
        return Database(dsets)

    def read_config(self):
        path = self.abs_path(self.config_path)
        kwargs = self.storage_kwargs(path)
        
        with fsspec.open(path, **kwargs) as f:
            config = yaml.safe_load(f)
        return config



from .hotspot import HotspotReport
from .anode_ramp import AnodeRampReport
from .anode_washing import AnodeWashingReport
from .abnormal_rates import AbnormalDAQRate
from .detector_operations import DetectorOperations



"""Attempt to register URConfig protocol if straxen
is installed.
"""
import straxen

import xedocs

DB_CACHE = {}

def get_accessor(name, db=None, **kwargs):
    if db is None:
        db = "straxen_db"
    db_func = getattr(xedocs.databases, db)
    db_kwargs = {k[4:] : v for k,v in kwargs.items() if k.startswith("db__")}
    db_key = (db, )
    if len(db_kwargs):
        db_key = db_key + tuple(sorted(db_kwargs.items()))
    if db_key not in DB_CACHE:
        DB_CACHE[db_key] = db_func(**db_kwargs)
    db = DB_CACHE[db_key]
    if name not in db:
        raise KeyError(f"{db_key} database has no such collection: {name}")
    return db[name]


@straxen.URLConfig.register("xedocs")
def xedocs_protocol(
    name, db=None, sort=None, attr=None, as_list=False, **labels
):
    """URLConfig protocol for fetching values from
        a xedocs database.
    ::param name: Name of the schema.
    ::param db: Database to load the data from.
    ::param version: Version of the documents to filter by.
    ::param sort: Attribute of the documents to sort on.
    ::param attr: Attribute of the documents to return.
    ::param labels: Label values to filter by to return.
    """

    accessor = get_accessor(name, db=db, **labels)

    kwargs = straxen.filter_kwargs(accessor.find_docs, labels)
    
    if isinstance(sort, (str,list)):
        kwargs["sort"] = sort

    if not as_list:
        kwargs["limit"] = 1

    docs = accessor.find_docs(**kwargs)

    if not docs:
        raise KeyError(
            f"No matching documents found for {name}. "
            "It is possible that there is no corresponding data."
        )

    if isinstance(sort, str):
        docs = sorted(docs, key=lambda x: getattr(x, sort))
    elif sort:
        docs = sorted(docs)

    if attr is not None:
        docs = [getattr(d, attr) for d in docs]

    if not as_list:
        return docs[0]

    return docs



import datetime
from typing import ClassVar, Literal
from pydantic import constr
import rframe

from xedocs import settings

from .base_corrections import BaseCorrectionSchema, TimeIntervalCorrection


class CorrectionReference(TimeIntervalCorrection):
    """A CorrectionReference document references one or
    more corrections by storing the name and labels required
    to locate the correction in a datasource
    """

    _ALIAS = ""

    # arbitrary alias for this reference,
    # this should match the straxen config name
    alias: str = rframe.Index(max_length=50)

    # the global version
    version: str = rframe.Index(max_length=20)

    # validity interval of the document
    time: rframe.Interval[datetime.datetime] = rframe.IntervalIndex()

    # Name of the correction being referenced
    correction: constr(max_length=50)

    # The attribute in the correction being referenced e.g `value`
    attribute: constr(max_length=50)

    # The index labels being referenced, eg pmt=[1,2,3], version='v3' etc.
    labels: dict

    def load(self, datasource=None, **overrides):
        """Load the referenced documents from the
        given datasource.
        """
        labels = dict(self.labels, **overrides)
        if self.correction not in BaseCorrectionSchema._CORRECTIONS:
            raise KeyError(f"Reference to undefined schema name {self.correction}")
        schema = BaseCorrectionSchema._CORRECTIONS[self.correction]
        return schema.find(datasource, **labels)

    @property
    def url_config(self):
        """Convert reference to a URLConfig URL"""
        import straxen

        url = f"{self.correction}://{self.attribute}"
        url = straxen.URLConfig.format_url_kwargs(url, **self.labels)
        return url

    @property
    def config_dict(self):
        return {self.name: self.url_config}


class BaseResourceReference(TimeIntervalCorrection):
    _ALIAS = ""

    fmt: ClassVar = "text"

    value: str

    def pre_insert(self, datasource):
        """require the existence of the resource
        being referenced prior to inserting a new
        document. This is to avoid typos etc.
        """
        self.load()
        super().pre_insert(datasource)

    def load(self, **kwargs):
        import straxen

        kwargs.setdefault("fmt", self.fmt)

        return straxen.get_resource(self.value, **kwargs)

    @property
    def url_config(self):
        return f"resource://{self.value}?fmt={self.fmt}"


class BaseMap(BaseResourceReference):
    _ALIAS = ""
    _ALLOWED_FORMATS = ['json.gz', 'json', 'tar.gz',
                        'tar', 'pkl.gz', 'pkl', 'npy',
                        'dill.gz','dill', 'npy_pickle',
                        'binary', 'text', 'txt', 'csv']

    algorithm: Literal["cnn", "gcn", "mlp", "cnf"] = rframe.Index()

    value: str

    @property
    def fmt(self):
        for fmt in self._ALLOWED_FORMATS:
            if self.value.endswith(fmt):
                break
        else:
            raise ValueError(f"Unknown format for {self.value}")
        return fmt

    @property
    def local_file(self):
        import utilix
        downloader = utilix.mongo_storage.MongoDownloader()
        return downloader.download_single(self.value)

    @property
    def file(self):
        import straxen
        file = straxen.get_resource(self.value, fmt=self.fmt)
        return file

    @property
    def map(self):
        import straxen
        return straxen.InterpolatingMap(self.file)


"""
Algorithm which uses a pretrained weightcube from the output of a 
Self-Organizing Map to classify peaklets as one of several SOM_types,
as well as classifying those types into either an S1 or an S2
"""

from ..base_references import BaseResourceReference


class SOMClassifier(BaseResourceReference):
    _ALIAS = 'som_classifiers'
    fmt = 'binary'
    
    value: str
    



import rframe
import datetime

from .base_pmt_data import BasePmtData
from ..constants import DETECTOR


class GainCalculation(BasePmtData):
    _ALIAS = "pmt_gain_calculations"

    detector: DETECTOR = rframe.Index()
    pmt: int = rframe.Index(ge=0)
    time: datetime.datetime = rframe.Index()

    voltage: float
    gain: float
    gain_error: float
    gain_stat_error: float
    gain_sys_error: float
    occupancy: float
    occupancy_error: float



from .utils import Database
from .xedocs import schemas_by_category, all_schemas
from .data_locations.mongodb import MongoDB
from .data_locations.corrections_repo import CorrectionsRepo
from .data_locations.api import XedocsApi
from .data_locations.data_folder import DataFolder


def straxen_db():
    schemas = all_schemas()
    db = MongoDB.from_utilix()
    accessors = {name: db.data_accessor(schema) for name, schema in schemas.items()}
    return Database(accessors)


def corrections_db():
    schemas = schemas_by_category()['corrections']
    db = MongoDB.from_utilix()
    accessors = {name: db.data_accessor(schema) for name, schema in schemas.items()}
    return Database(accessors)


def corrections_repo(branch=None, username=None, token=None):
    kwargs = {}
    if branch is not None:
        kwargs["branch"] = branch
    if username is not None:
        kwargs["username"] = username
    if token is not None:
        kwargs["token"] = token
    repo = CorrectionsRepo(**kwargs)
    return repo.get_datasets()


def xedocs_api(login=True):
    api = XedocsApi()
    if login:
        api.login()
    return api.get_datasets()


def default_db():
    return straxen_db()


def development_db():
    schemas = all_schemas()
    db = MongoDB.from_utilix(db_name='xedocs-dev')
    accessors = {name: db.data_accessor(schema) for name, schema in schemas.items()}
    return Database(accessors)


def local_mongo_db(**kwargs):
    schemas = all_schemas()
    db = MongoDB(**kwargs)
    accessors = {name: db.data_accessor(schema) for name, schema in schemas.items()}
    return Database(accessors)


def local_folder(path: str = None, **kwargs):
    if path is None:
        raise ValueError("Path must be set")
    repo = DataFolder(root=path, **kwargs)
    return repo.get_datasets()


import pydantic
import datetime
from typing import Literal, List
from .base_report import BaseOperationsReport


class LevelChange(pydantic.BaseModel):
    time: datetime.datetime
    liquid_level: float = pydantic.Field(lt=8, ge=0)


class AnodeWashingReport(BaseOperationsReport):
    """Anode washing report"""

    _ALIAS = "anode_washes"

    washes: List[LevelChange]



import os
from typing import Any, Dict
from pydantic import BaseSettings
from rframe import DataAccessor


def xenon_config_source(settings: BaseSettings) -> Dict[str, Any]:
    from xedocs import settings as xenon_settings

    cfg = xenon_settings.xenon_config.RunDB
    data = {}
    if cfg.xent_url:
        data['host'] = cfg.xent_url
    if cfg.xent_user:
        data['username'] = cfg.xent_user
    if cfg.xent_password:
        data['password'] = cfg.xent_password
    if cfg.xent_database:
        data['auth_db'] = cfg.xent_database
    if cfg.max_pool_size:
        data['max_pool_size'] = cfg.max_pool_size
    if cfg.socket_timeout:
        data['socket_timeout'] = cfg.socket_timeout
    if cfg.connect_timeout:
        data['connect_timeout'] = cfg.connect_timeout
    if cfg.read_preference:
        data['read_preference'] = cfg.read_preference
    return data


class MongoDB(BaseSettings):
    CLIENT_CACHE = {}

    class Config:
        env_prefix = "xedocs_mongo_"
        secrets_dir = '/run/secrets'

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            sources = (
                init_settings,
                env_settings,
                xenon_config_source,
            )
            if os.path.isdir("/run/secrets"):
                sources = sources + (file_secret_settings,)
            return sources
        
    username: str = None
    password: str = None
    db_name: str = "xedocs"
    auth_db: str = "admin"
    host: str = "localhost"
    connection_uri: str = None

    max_pool_size: int = 100
    socket_timeout: int = 60000
    connect_timeout: int = 60000
    read_preference: str = "secondaryPreferred"

    @property
    def client(self):
        if self.connection_uri is None:
            self.connection_uri = f"mongodb://{self.username}:{self.password}@{self.host}?authSource={self.auth_db}"
        if self.connection_uri not in self.CLIENT_CACHE:
            self.CLIENT_CACHE[self.connection_uri] = self.make_client(self.connection_uri)
        return self.CLIENT_CACHE[self.connection_uri]

    def make_client(self, connection_uri):
        import pymongo
        return pymongo.MongoClient(connection_uri, 
                                   maxPoolSize=self.max_pool_size,
                                   socketTimeoutMS=self.socket_timeout,
                                   connectTimeoutMS=self.connect_timeout,
                                   readPreference=self.read_preference)

    @classmethod
    def from_utilix(cls, **kwargs):
        from utilix import uconfig
        host = uconfig.get('RunDB', 'xent_url')
        username = uconfig.get('RunDB', 'xent_user')
        password = uconfig.get('RunDB', 'xent_password')
        auth_db = uconfig.get('RunDB', 'xent_database')
        config = dict(host=host, username=username,
                           password=password, auth_db=auth_db)
        config.update(kwargs)
        return cls(**config)

    def data_accessor(self, schema):
        datasource = self.client[self.db_name][schema._ALIAS]
        return DataAccessor(schema, datasource)



from typing import Dict

import rframe

from .base_schemas import VersionedXeDoc
from .plugin_lineages import PluginLineage


class ContextLineage(VersionedXeDoc):
    _ALIAS = "context_lineages"

    strax: str = rframe.Index()
    straxen: str = rframe.Index()

    lineage_hashes: Dict[str, str]

    def load_config(self, datasource=None):
        hashes = list(self.lineage_hashes.values())
        docs = PluginLineage.find(datasource, lineage_hash=hashes)
        config = {}
        for doc in docs:
            config.update(doc.config)
        return config



# This will be a correction designed for testing purposes.
# This SHOULD NOT be used in production.

import rframe

from ..base_corrections import TimeSampledCorrection
from ...constants import DETECTOR

class TestCorrection(TimeSampledCorrection):
    _ALIAS = "test_corrections"

    # Correction identical to PMT gains.
    # Used for testing

    detector: DETECTOR = rframe.Index()
    pmt: int = rframe.Index(ge=0)

    value: float
